import torch


def select_dataset(dataloader_type="worldstrat_preprocessed_RGBNIR",train_batch_size=2,val_batch_size=2,num_workers=0,revisits=1,prefetch_factor=0,**kwargs):
    """
    Creates training and validation data loaders from a specified dataset.

    The function supports multiple datasets and the returned data loaders
    are ready for use in a training loop.

    Args:
        dataloader_type (str, optional): Specifies the type of dataset to use. Options include "worldstrat_preprocessed", 
        "NAIP", "worldstrat_preprocessed_RGBNIR", "coco_dataset", "worldstrat_spot6", and "s2_dataset". 
        Default is "worldstrat_preprocessed_RGBNIR".

        train_batch_size (int, optional): Number of samples per batch to load for training data. Default is 2.

        val_batch_size (int, optional): Number of samples per batch to load for validation data. Default is 2.

        num_workers (int, optional): Number of subprocesses to use for data loading. Default is 1.

    Returns:
        tuple: Consisting of (train_loader, val_loader, ds_train, ds_val) where train_loader and val_loader are 
        torch.utils.data.DataLoader instances for the training and validation datasets respectively, and ds_train, ds_val
        are the corresponding dataset instances.
    """
    
    # get and control datalaoder
    assert dataloader_type in ["S2NAIP_v4","sen2_test"]

    # Each conditional block sets up data loaders for a different dataset
    # Loaders are configured according to specific dataset require
    
    if dataloader_type == "S2NAIP_v4":
        import pathlib
        from opensr_dataloaders.S2NAIP_v4 import S2NAIP_v4
        from torch.utils.data import DataLoader
        csv_path = "/data3/landcover_s2naip/csvs/train_metadata_landcover.csv"
        ds_train = S2NAIP_v4(phase="train",csv_path=csv_path)
        ds_val =  S2NAIP_v4(phase="val",csv_path=csv_path)
        train_loader = torch.utils.data.DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,
                                                   drop_last=True,prefetch_factor=prefetch_factor,
                                                   sampler = ds_train.sampler)
        val_loader = torch.utils.data.DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,
                                                 drop_last=True,prefetch_factor=prefetch_factor,)
                                                 #sampler = ds_val.sampler)

    if dataloader_type == "worldstrat_MISR":
        import pathlib
        from ldm.data.worldstrat_MISR import worldstrat_MISR
        from torch.utils.data import DataLoader
        revisits = 4
        csv_path = "/data3/landcover_s2naip/csvs/train_metadata_landcover.csv"
        ds_train = worldstrat_MISR(root_dir="/data3/worldstrat_misr_saved_2",phase="train",revisits=revisits,stddev_noise=0.1,noise=False)
        ds_val =  worldstrat_MISR(root_dir="/data3/worldstrat_misr_saved_2",phase="val",revisits=revisits,stddev_noise=0.1,noise=False)
        train_loader = torch.utils.data.DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,
                                                   drop_last=True,prefetch_factor=prefetch_factor)
        val_loader = torch.utils.data.DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,
                                                 drop_last=True,prefetch_factor=prefetch_factor)

    if dataloader_type == "S2NAIP":
        import pathlib
        from ldm.data.S2NAIP import S2NAIP
        from torch.utils.data import DataLoader
        ds_train = S2NAIP(phase="train",data_folder="/data2/simon/S2NAIP_saved/train/",apply_norm=True)
        ds_val =  S2NAIP(phase="val",data_folder="/data2/simon/S2NAIP_saved/train/",apply_norm=True)
        train_loader = torch.utils.data.DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,
                                                   shuffle=True,drop_last=True,prefetch_factor=prefetch_factor)
        val_loader = torch.utils.data.DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,
                                                 drop_last=True,shuffle=True,prefetch_factor=prefetch_factor)
    
    if dataloader_type == "S2NAIP_v2":
        import pathlib
        from ldm.data.S2NAIP_v2 import S2NAIP_v2
        from torch.utils.data import DataLoader
        ds_train = S2NAIP_v2(phase="train")
        ds_val =  S2NAIP_v2(phase="val")
        train_loader = torch.utils.data.DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,
                                                   shuffle=True,drop_last=True,prefetch_factor=prefetch_factor)
        val_loader = torch.utils.data.DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,
                                                 drop_last=True,shuffle=True,prefetch_factor=prefetch_factor)
        
    if dataloader_type == "S2NAIP_v2_MISR":
        import pathlib
        from ldm.data.S2NAIPv2_MISR_fake import S2NAIPv2_MISR_fake
        from torch.utils.data import DataLoader
        ds_train = S2NAIPv2_MISR_fake(phase="train")
        ds_val =  S2NAIPv2_MISR_fake(phase="val")
        train_loader = torch.utils.data.DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,
                                                   shuffle=True,drop_last=True,prefetch_factor=prefetch_factor)
        val_loader = torch.utils.data.DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,
                                                 drop_last=True,shuffle=True,prefetch_factor=prefetch_factor)

    if dataloader_type == "S2NAIP_v2_20m":
        import pathlib
        from ldm.data.S2NAIP_v2_6b_20m import S2NAIP_v2_6b_20m
        from torch.utils.data import DataLoader
        ds_train = S2NAIP_v2_6b_20m(phase="train",data_folder = "/data2/simon/S2NAIP_v2_b6_20m/",apply_norm=True,augmentation=True)
        ds_val =  S2NAIP_v2_6b_20m(phase="val",data_folder = "/data2/simon/S2NAIP_v2_b6_20m/",apply_norm=True,augmentation=True)
        train_loader = torch.utils.data.DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,
                                                   shuffle=True,drop_last=True,prefetch_factor=prefetch_factor)
        val_loader = torch.utils.data.DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,
                                                 drop_last=True,shuffle=False,prefetch_factor=prefetch_factor)
        
    
    if dataloader_type == "sen2_test":
        import pathlib
        from opensr_dataloaders.sen2_test import sen2_test
        from torch.utils.data import DataLoader
        ds_test = sen2_test(data_folder = "/data2/simon/test_s2/S2A_MSIL2A_20230729T100031_N0509_R122_T33TUG_20230729T134559.SAFE",
                            amount=24,band_selection=kwargs["band_selection"],apply_norm=kwargs["apply_norm"])
        test_loader = torch.utils.data.DataLoader(ds_test,batch_size=val_batch_size,num_workers=num_workers,
                                                   shuffle=False,drop_last=True,prefetch_factor=prefetch_factor)
        return(test_loader)
        
    if dataloader_type == "S2NAIP_int":
        import pathlib
        from ldm.data.S2NAIP_int import S2NAIP_int
        from torch.utils.data import DataLoader
        ds_train = S2NAIP_int(phase="train",data_folder="/data2/simon/S2NAIP_saved/train/")
        ds_val =  S2NAIP_int(phase="val",data_folder="/data2/simon/S2NAIP_saved/train/")
        train_loader = torch.utils.data.DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,
                                                   shuffle=True,drop_last=True,prefetch_factor=prefetch_factor)
        val_loader = torch.utils.data.DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,
                                                 drop_last=True,shuffle=True,prefetch_factor=prefetch_factor)

    if dataloader_type == "worldstrat_spot6":
        from ldm.data.custom_data import ImageDataset
        from torch.utils.data import DataLoader

        dataset_root = "/data2/simon/worldstrat_SISR_preprocessed/"
        ds_train = ImageDataset(phase="train",dataset_type="worldstrat",return_type="pair",
                            lr_res=128,hr_res=256,return_res="interpolated",val_range="0..1")
        ds_val = ImageDataset(phase="val",dataset_type="worldstrat",return_type="pair",
                            lr_res=128,hr_res=256,return_res="interpolated",val_range="0..1")

        train_loader = DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,shuffle=True,drop_last=True)
        val_loader = DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,drop_last=True)

    if dataloader_type == "worldstrat_preprocessed":
        from ldm.data.custom_data_worldstrat_preprocessed import worldstrat_SISR
        from torch.utils.data import DataLoader

        dataset_root = "/data2/simon/worldstrat_SISR_preprocessed/"
        ds_train = worldstrat_SISR(dataset_root,phase="train")
        ds_val = worldstrat_SISR(dataset_root,phase="val")

        train_loader = DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,shuffle=True,drop_last=True)
        val_loader = DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,drop_last=True,shuffle=True)

    if dataloader_type == "worldstrat_preprocessed_RGBNIR":
        from ldm.data.custom_data_worldstrat_preprocessed import worldstrat_SISR
        from torch.utils.data import DataLoader

        dataset_root = "/data2/simon/worldstrat_SISR_RGBNIR_preprocessed/"
        ds_train = worldstrat_SISR(dataset_root,phase="train",lr_type="interpolated")
        ds_val = worldstrat_SISR(dataset_root,phase="val",lr_type="interpolated",)

        train_loader = DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,drop_last=True,shuffle=True)
        val_loader = DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,drop_last=True,shuffle=True)


    if dataloader_type == "s2_dataset":
        from opensr_dataloaders.custom_sen2 import s2_dataset
        from torch.utils.data import DataLoader
        #directory_path = '/data3/S2_images/'  # Replace with the actual directory path
        directory_path = '/data2/simon/S2_images/'  # Replace with the actual directory path 
        ds_train = s2_dataset(directory_path,phase="train",bands=4)
        ds_val = s2_dataset(directory_path,phase="val",bands=4)
        train_loader = DataLoader(ds_train, batch_size=train_batch_size, drop_last=True,shuffle=True,num_workers=num_workers)
        val_loader = DataLoader(ds_val, batch_size=val_batch_size, drop_last=True,shuffle=True,num_workers=num_workers)
        
    if dataloader_type == "coco_dataset":
        from ldm.data.coco_dataset import JpegImageDataset as coco_dataset
        from torch.utils.data import DataLoader

        folders = ["/data2/simon/cv_datasets/COCO","/data2/simon/cv_datasets/ADE20K","/data2/simon/cv_datasets/OpenImages"]
        ds_train = coco_dataset(folders=folders,phase="train",image_size=512,return_type="pair",channels=4,force_rescan=False)
        ds_val = coco_dataset(folders=folders,phase="val",image_size=512,return_type="pair",channels=4,force_rescan=False)

        train_loader = DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,shuffle=True,persistent_workers=True)
        val_loader = DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,persistent_workers=True,drop_last=True,shuffle=True)

    if dataloader_type == "worldstrat_MISR_fake":
        from ldm.data.worldstrat_MISR_fake import worldstrat_MISR_fake as worldstrat_MISR_fake
        from torch.utils.data import DataLoader
        
        ds_train = worldstrat_MISR_fake(phase="train",revisits=revisits,noise=False)
        ds_val = worldstrat_MISR_fake(phase="val",revisits=revisits,noise=False)
        
        train_loader = DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,shuffle=True,persistent_workers=True)
        val_loader = DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,persistent_workers=True,drop_last=True,shuffle=True)


    if dataloader_type == "worldstrat_r":
        from ldm.data.worldstrat_r import worldstrat_DataSet
        from torch.utils.data import DataLoader
        
        ds_train = worldstrat_DataSet("/data1/worldstrat_r",ftype="train",revisits=revisits)
        ds_val = worldstrat_DataSet("/data1/worldstrat_r",ftype="val",revisits=revisits)
        
        train_loader = DataLoader(ds_train,batch_size=train_batch_size,num_workers=num_workers,shuffle=True,persistent_workers=True)
        val_loader = DataLoader(ds_val,batch_size=val_batch_size,num_workers=num_workers,persistent_workers=True,drop_last=True,shuffle=True)

        
        
        
    # FINAL RETURN
    return train_loader,val_loader,ds_train,ds_val
        


def create_pl_datamodule(train_loader,val_loader,test_loader=None):
    import pytorch_lightning as pl
    class pl_datamodule(pl.LightningDataModule):
        def __init__(self, train_loader, val_loader):
            super().__init__()
            self.train_loader = train_loader
            self.val_loader = val_loader
            self.test_loader = test_loader
        def train_dataloader(self):
            return self.train_loader
        def val_dataloader(self):
            return self.val_loader
        def test_dataloader(self):
            if self.test_loader!=None:
                return self.test_loader
        def prepare_data(self):
            pass
        def setup(self, stage=None):
            pass
        

    datamodule = pl_datamodule(train_loader,val_loader)
    return(datamodule)


"""
Copy of prev method in case sth goes wrong

# get and control datalaoder
dataloader_type = "worldstrat_preprocessed_RGBNIR"

batch_size = 2

assert dataloader_type in ["worldstrat_preprocessed","NAIP",
                           "worldstrat_preprocessed_RGBNIR",
                           "coco_dataset","worldstrat_spot6",]


if dataloader_type == "NAIP":
    from ldm.data.custom_data_naip import NAIP
    from torch.utils.data import DataLoader
    ds_train = NAIP(pathlib.Path("/data3/S2NAIP/train"), type="train", scale=4)
    ds_val =  NAIP(pathlib.Path("/data3/S2NAIP/val"), type="val", scale=4)
    ds_val = torch.utils.data.Subset(ds_val,list(range(20)))
    train_loader = torch.utils.data.DataLoader(ds_train,batch_size=16,shuffle=True)
    val_loader = torch.utils.data.DataLoader(ds_val,batch_size=5,drop_last=True)

if dataloader_type == "worldstrat_spot6":
    from ldm.data.custom_data import ImageDataset
    from torch.utils.data import DataLoader
    
    dataset_root = "/data2/simon/worldstrat_SISR_preprocessed/"
    ds_train = ImageDataset(phase="train",dataset_type="worldstrat",return_type="pair",
                        lr_res=128,hr_res=256,return_res="interpolated",val_range="0..1")
    ds_val = ImageDataset(phase="val",dataset_type="worldstrat",return_type="pair",
                        lr_res=128,hr_res=256,return_res="interpolated",val_range="0..1")
    
    train_loader = DataLoader(ds_train,batch_size=batch_size,num_workers=16,shuffle=True)
    val_loader = DataLoader(ds_val,batch_size=10,num_workers=batch_size)

if dataloader_type == "worldstrat_preprocessed":
    from ldm.data.custom_data_worldstrat_preprocessed import worldstrat_SISR
    from torch.utils.data import DataLoader
    
    dataset_root = "/data2/simon/worldstrat_SISR_preprocessed/"
    ds_train = worldstrat_SISR(dataset_root,phase="train")
    ds_val = worldstrat_SISR(dataset_root,phase="val")
    
    train_loader = DataLoader(ds_train,batch_size=batch_size,num_workers=16,shuffle=True)
    val_loader = DataLoader(ds_val,batch_size=10,num_workers=batch_size)
    
if dataloader_type == "worldstrat_preprocessed_RGBNIR":
    from ldm.data.custom_data_worldstrat_preprocessed import worldstrat_SISR
    from torch.utils.data import DataLoader
    
    dataset_root = "/data2/simon/worldstrat_SISR_RGBNIR_preprocessed/"
    ds_train = worldstrat_SISR(dataset_root,phase="train",lr_type="interpolated")
    ds_val = worldstrat_SISR(dataset_root,phase="val",lr_type="interpolated",)
    
    train_loader = DataLoader(ds_train,batch_size=batch_size,num_workers=16,shuffle=True)
    val_loader = DataLoader(ds_val,batch_size=batch_size,num_workers=1)
    
    
if dataloader_type == "coco_dataset":
    from ldm.data.coco_dataset import JpegImageDataset as coco_dataset
    from torch.utils.data import DataLoader
    
    folders = ["/data2/simon/cv_datasets/COCO","/data2/simon/cv_datasets/ADE20K","/data2/simon/cv_datasets/OpenImages"]
    ds_train = coco_dataset(folders=folders,phase="train",image_size=512,return_type="single",channels=4,force_rescan=False)
    ds_val = coco_dataset(folders=folders,phase="val",image_size=512,return_type="single",channels=4,force_rescan=False)
    
    train_loader = DataLoader(ds_train,batch_size=batch_size,num_workers=16,shuffle=True,persistent_workers=True)
    val_loader = DataLoader(ds_val,batch_size=1,num_workers=1,persistent_workers=True)
"""



import os
import torch
from PIL import Image
import numpy as np
import os
import glob
from torch.utils.data import Dataset, DataLoader
from einops import rearrange
import torchvision.transforms as transforms

import json
from einops import rearrange
from torchvision import transforms

class JpegImageDataset(Dataset):
    def __init__(self,folders,phase,image_size=512,return_type="single",channels=3,force_rescan=False,verbose=False):
        
        self.verbose=verbose
        
        # set inputs to class
        self.channels = channels
        self.folders = folders
        self.image_size = image_size # desired image dimensions
        self.return_type = return_type
        self.phase = phase
        
        # set transforms
        self.init_transforms(dataset_type="coco_dataset",
                             band_stats_path="ldm/data/band_statistics.json",
                             apply_transforms=False)

        # make sure inputs are valid
        assert self.channels in [3,4]
        assert self.return_type in ["single","pair"]
        assert self.phase in ["train","val"]
        assert len(self.folders)>0
        
        if self.verbose:
            print("Creating dataset from all png/jpeg files in directories:",self.folders,"\n")
                
        # Read files from folders, by:
        # 1. rescanning folders and check for image size & RGB
        # 2. use list with precomputed info
        self.image_files = self.read_image_file_or_scan_folders(self.folders,force_rescan=force_rescan)
        if self.verbose:
            print("Total number of suitable image files in all folders: ",len(self.image_files))
        
        # train/test split
        self.image_files = self.train_test_split(phase)
        if self.verbose:
            print("No. of images after filtering for",self.phase,"pase: ",len(self.image_files),"\n")
            print("Dataset sucessfully initialized.")
        
    
    def train_test_split(self,phase="train"):
        # perform train test split on data
        if phase=="train":
            #first_85_percent = self.image_files[:int(len(self.image_files)*0.85)]
            #image_files = first_85_percent
            image_files = self.image_files[:len(self.image_files)-5]
            #image_files = self.image_files[:100]
        if phase=="val":
            #last_15_percent = self.image_files[-int(len(self.image_files)*0.15):]
            #image_files = last_15_percent
            image_files = self.image_files[len(self.image_files)-5:]
        return(image_files)
    
    def read_image_file_or_scan_folders(self,folders,force_rescan=False):
        list_of_lists = []
        for folder in folders:
            if self.verbose:
                print("Loading from folder:",folder)
            # first, check if folder has been scanned for suitable pictures already
            suitable_pictures_file_path = os.path.join(folder, 'suitable_files.txt')
            if os.path.exists(suitable_pictures_file_path) and force_rescan==False:
                if self.verbose:
                    print('List containing suitable files found (or force_rescan is activated):',suitable_pictures_file_path)
            
                with open(suitable_pictures_file_path, 'r') as f:
                    image_files_from_folder = [line.strip() for line in f.readlines()]
            else:
                if self.verbose:
                    print('No list containing suitable images found. Scanning and creating ',suitable_pictures_file_path)
                # read all file names from folder
                image_files_from_folder = self.get_images_from_folders(folder)
                # check each image to see if they are RGB and over the size requirement
                image_files_from_folder = self.check_image_sizes(image_files_from_folder)
                # save suitable files list to text file in root directory of dataset
                with open(suitable_pictures_file_path, 'w') as f:
                    for item in image_files_from_folder:
                        f.write("%s\n" % item)
        
            
            # append list of files either loaded or scanned to self         
            list_of_lists.append(image_files_from_folder)
            if self.verbose:
                print("No. of suitable image files in this folder:",len(image_files_from_folder),"\n")
        return(self.flatten_list(list_of_lists))
            
    
    def get_images_from_folders(self,folder_path):
        image_files = glob.glob(os.path.join(folder_path, '**/*.jpg'), recursive=True)
        image_files = self.check_image_sizes(image_files)
        return(image_files)
    
    def check_image_sizes(self,image_files):
        new_ls = []
        for path in image_files:
            with Image.open(path) as image:
                width, height = image.size
                if width>self.image_size and height>self.image_size and image.mode=="RGB":
                    new_ls.append(path)
        return(new_ls)
    
    def crop_center(self,image):
        height, width = image.shape[:2]
        crop_size = self.image_size
        y_start = (height - crop_size) // 2
        x_start = (width - crop_size) // 2
        y_end = y_start + crop_size
        x_end = x_start + crop_size
        return image[y_start:y_end, x_start:x_end]
    
    def flatten_list(self,nested_list):
        flat_list = []
        for item in nested_list:
            if isinstance(item, list):
                flat_list.extend(self.flatten_list(item))
            else:
                flat_list.append(item)
        return flat_list
    
    def rgb_to_grayscale(self,img_tensor):
        r, g, b = img_tensor[0], img_tensor[1], img_tensor[2]
        gray = 0.2989 * r + 0.5870 * g + 0.1140 * b  # grayscale formula
        return torch.stack([r, g, b, gray], dim=0)

        
    def __len__(self):
        return len(self.image_files)

    def init_transforms(self,dataset_type="worldstrat_preprocessed_RGBNIR",band_stats_path="ldm/data/band_statistics.json",apply_transforms=True):

        
        if apply_transforms==False:
            self.transform_PIL = transforms.ToPILImage()
            self.transform = transforms.ToTensor()
        
        if apply_transforms==True:
            try:
                with open(band_stats_path, 'r') as file:
                    # Load the JSON data into a dictionary
                    data_dict = json.load(file)
                mean = [data_dict["norm_dict"][dataset_type]["mean"]["0"],
                        data_dict["norm_dict"][dataset_type]["mean"]["1"],
                        data_dict["norm_dict"][dataset_type]["mean"]["2"],
                        data_dict["norm_dict"][dataset_type]["mean"]["3"]]

                std = [data_dict["norm_dict"][dataset_type]["std"]["0"],
                        data_dict["norm_dict"][dataset_type]["std"]["1"],
                        data_dict["norm_dict"][dataset_type]["std"]["2"],
                        data_dict["norm_dict"][dataset_type]["std"]["3"]]
                transform_ = transforms.Compose([
                        transforms.ToTensor(),
                        transforms.Normalize(mean=mean, std=std),
                         ])
                self.transform_PIL = transforms.ToPILImage()
                self.transform = transform_

            except:
                self.transform_PIL = transforms.ToPILImage()
                self.transform = transforms.ToTensor()

    def apply_transforms(self,im):
        """
        in: - torch tensor of shape W H C
        out: normalized tensor
        """
        # check if we need transpose at this place
        if im.shape[-1]<im.shape[0]: # True if W H B
            transpose_needed = True
        else:
            transpose_needed = False
        
        # apply transpose if needed, then apply transform
        if transpose_needed:
            im = rearrange(im,"w h b -> b w h")
        im = self.transform_PIL(im)
        im = self.transform(im)
        if transpose_needed:
            im = rearrange(im,"b w h -> w h b")
        return(im)
    
    
    def __getitem__(self, index):
        image_path = self.image_files[index]
        image = Image.open(image_path)
        image = np.array(image)
        image = torch.Tensor(image)
        
        image = self.crop_center(image)
        image = image/255.0
        
        # get 4th band from grayscale
        if self.channels==4:
            # turn to RGB if wanted
            image = rearrange(image,"w h c -> c w h")
            image = self.rgb_to_grayscale(image)
            image = rearrange(image,"c w h -> w h c")
            # apply transformation/normalization
            image = self.apply_transforms(image)
        
        # get interpolated version and return
        if self.return_type =="pair":
            # rearrange hr image
            image = rearrange(image,"w h c -> c w h")
            # interpolate hr to lr
            image_lr = torch.nn.functional.interpolate(image.unsqueeze(0),
                                                    size=(128,128), mode='bilinear', align_corners=False).squeeze(0)
            # bring images back to wanted shape
            image    = rearrange(image,"c w h -> w h c")
            image_lr = rearrange(image_lr,"c w h -> w h c")

            # stretch to -1..+1
            image = (image*2)-1
            image_lr = (image_lr*2)-1

            # return images as dict
            return {"image":image,"LR_image":image_lr}
            
        if self.return_type =="single":
            # stretch to -1..+1
            #image = (image*2)-1
            return {"image":image}


import torch
import os
import einops
import rasterio
from rasterio.windows import Window
from torchvision import transforms
import random
import json
import numpy as np


class sen2_test(torch.utils.data.Dataset):
    def __init__(self,data_folder = "/data2/simon/test_s2/S2A_MSIL2A_20230729T100031_N0509_R122_T33TUG_20230729T134559.SAFE",amount=25,band_selection="R10m",apply_norm=False):
        # settings for band selection
        assert band_selection in ["R10m","R20m"]
        self.band_selection = band_selection

        # init transforms
        self.apply_norm=apply_norm
        
        # def window return size
        self.image_size=128
        # get location of image data
        for dirpath, dirnames, _ in os.walk(data_folder):
            if "IMG_DATA" in dirnames:
                folder_path = os.path.join(dirpath, "IMG_DATA")
        folder_path = os.path.join(folder_path,self.band_selection)
        file_paths = os.listdir(folder_path)

        # get image file paths for selected bands
        if self.band_selection == "R10m":
            self.image_files = {"R":os.path.join(folder_path,[file for file in file_paths if "B04" in file][0]),
                        "G":os.path.join(folder_path,[file for file in file_paths if "B03" in file][0]),
                        "B":os.path.join(folder_path,[file for file in file_paths if "B02" in file][0]),
                        "NIR":os.path.join(folder_path,[file for file in file_paths if "B08" in file][0])}
        if self.band_selection == "R20m":
            self.image_files = {"B05":os.path.join(folder_path,[file for file in file_paths if "B05" in file][0]),
                        "B06":os.path.join(folder_path,[file for file in file_paths if "B06" in file][0]),
                        "B07":os.path.join(folder_path,[file for file in file_paths if "B07" in file][0]),
                        "B8A":os.path.join(folder_path,[file for file in file_paths if "B8A" in file][0]),
                        "B11":os.path.join(folder_path,[file for file in file_paths if "B11" in file][0]),
                        "B12":os.path.join(folder_path,[file for file in file_paths if "B12" in file][0])}
        
        # extract keys from image files
        self.band_names = list(self.image_files.keys())
        
        # get iamge shape
        with rasterio.open(self.image_files[self.band_names[0]]) as src:
            self.image_width = src.width
            self.image_height = src.height
        
        # create list of coordinates
        self.windows = []
        for i in range(amount):
            rand_x = random.randint(0 ,self.image_width -self.image_size)
            rand_y = random.randint(0,self.image_height-self.image_size)
            window_ = Window(rand_x, rand_y, self.image_size, self.image_size)
            self.windows.append(window_)

    def __len__(self):
        return len(self.windows)
    
    def apply_transforms(self,im):
        from opensr_dataloaders.linear_transforms import linear_transform
        im = linear_transform(im.unsqueeze(0),stage="norm").squeeze(0)
        #im = einops.rearrange(im,"C W H -> W H C")
        return(im)
    
    def __getitem__(self,idx):
        # get current window
        window = self.windows[idx]
        
        # read bands iteratively
        image=[]
        for band in self.band_names:
            with rasterio.open(self.image_files[band]) as src:
                window_data = src.read(1, window=window)                
                image.append(window_data)
        image = np.stack(image)/10000
        
        # apply normalization
        if self.apply_norm:
            image = self.apply_transforms(torch.Tensor(image))

        image = torch.Tensor(image).float()
        return(image)
                

"""
ds = sen2_test(band_selection="R20m",apply_norm=False)
im = ds.__getitem__(20)
import matplotlib.pyplot as plt
plt.figure(figsize=(10, 6))
colors = ['b', 'g', 'r', 'c', 'm', 'y']
num_bins = 10
# print min, mean, std,max for all image bands
for i in range(im.shape[0]):
    band = im[i]
    min_val = band.min().item()
    mean_val = band.mean().item()
    std_val = band.std().item()
    max_val = band.max().item()
    print(f"Band {i+1}: Min={min_val}, Mean={mean_val}, Std={std_val}, Max={max_val}")
    band = im[i].flatten()
    plt.hist(band, bins=30, color=colors[i], alpha=0.5, label=f'Band {ds.band_names[i]}')
    # plot line at max value
    plt.axvline(x=band.max(), color=colors[i], linestyle='dashed', linewidth=1)
    #band = im[i].flatten().numpy()  # Convert to NumPy array for histogram calculation
    #hist, bin_edges = np.histogram(band, bins=num_bins)
    #bin_centers = 0.5 * (bin_edges[:-1] + bin_edges[1:])  # Calculate bin centers
    #plt.plot(bin_centers, hist, color=colors[i], label=f'Band {ds.band_names[i]}')
plt.title('Histograms of Each Band')
plt.xlabel('Value')
plt.ylabel('Frequency')
plt.legend()
plt.savefig("hist.png")
plt.close()
"""


import os
from PIL import Image
import torch
from torch.utils.data import Dataset, DataLoader
import imageio.v2 as imageio
import cv2
import numpy as np

import json
from einops import rearrange
from torchvision import transforms

class s2_dataset(Dataset):
    def __init__(self, directory,phase="train",bands=4):
        # set properties and directories
        self.directory = directory
        self.lr_path = os.path.join(self.directory,"lr")
        self.hr_path = os.path.join(self.directory,"hr")
        self.bands = bands
        self.phase = phase
        
        # read files
        self.file_list_hr = os.listdir(self.hr_path)
        self.file_list_lr = os.listdir(self.lr_path)
        # keep only files in both lists
        self.files = list(set(self.file_list_hr).intersection(self.file_list_lr))
        
        # apply phase filtering
        phase_percentages = {"train":0.995,"val":0.0025,"test":0.0025}
        total_images = len(self.files)
        train_count = int(total_images * phase_percentages["train"])
        val_count = int(total_images * phase_percentages["val"])
        test_count = int(total_images * phase_percentages["test"])

        # Split the list into train, validation, and test sets
        if phase == "train":
            self.files = self.files[:train_count]
        if phase == "val":
            self.files = self.files[train_count:train_count + val_count]
        if phase == "test":
            self.files = self.files[train_count + val_count:]
            
            
            
    def __len__(self):
        return len(self.files)
    
    def linear_transform(self,t,stage="norm"):
        assert stage in ["norm","denorm"]    

        # define constants
        self.rgb_c = 3.
        self.nir_c = 5.

        if stage == "norm":
            # divide according to conventions
            t[:,:,0] = t[:,:,0] * (10.0 / self.rgb_c) # R
            t[:,:,1] = t[:,:,1] * (10.0 / self.rgb_c) # G
            t[:,:,2] = t[:,:,2] * (10.0 / self.rgb_c) # B
            t[:,:,3] = t[:,:,3] * (10.0 / self.nir_c) # NIR
            # clamp to get rid of outlier pixels
            t = t.clamp(0,1)
            # bring to -1..+1
            t = (t*2)-1
        if stage == "denorm":
            # bring to 0..1
            t = (t+1)/2
            # divide according to conventions
            t[:,:,0] = t[:,:,1] * (self.rgb_c / 10.0) # R
            t[:,:,1] = t[:,:,1] * (self.rgb_c / 10.0) # G
            t[:,:,2] = t[:,:,2] * (self.rgb_c / 10.0) # B
            t[:,:,3] = t[:,:,3] * (self.nir_c / 10.0) # NIR
            # clamp to get rif of outlier pixels
            t = t.clamp(0,1)

        # return image
        return(t)


    def apply_augmentations(self,im):
        smoothen_liklelihood = 1.0
        jitter_liklelihood   = 0.0
        
        if self.phase=="train" or self.phase=="val":
            import random
            smoothen_rand = random.uniform(0, 1)
            jitter_rand = random.uniform(0, 1)

            # denormalize and rearrange
            im = (im+1)/2
            im = rearrange(im,"w h c -> c w h")

            # perform color jitter
            if jitter_rand<jitter_liklelihood:
                # Define the color jitter transformation
                color_jitter = transforms.ColorJitter(brightness=0.5, contrast=0.3, saturation=0.3, hue=0.3)
                # Apply the transformation
                band_ls = []
                for band in im:
                    band = torch.unsqueeze(band,0)
                    #print(band.shape)
                    band = color_jitter(band)
                    band = torch.squeeze(band,0)
                    band_ls.append(band)
                im = torch.stack(band_ls)
            
            # perform smoothen
            if smoothen_rand<smoothen_liklelihood:
                # Define Kernel
                sigma_rand = random.uniform(0.7, 1.0)
                sigma_rand = 0.4
                gaussian_blur = transforms.GaussianBlur(kernel_size=5, sigma=sigma_rand)
                # Apply the blur to the image tensor
                band_ls = []
                for band in im:
                    band = torch.unsqueeze(band,0)
                    band = gaussian_blur(band)
                    band = torch.squeeze(band,0)
                    band_ls.append(band)
                im = torch.stack(band_ls)
                # perform jitter

            # normalize and rearrange
            im = (im*2)-1
            im = rearrange(im,"c w h -> w h c")
        
        # return im regardless of train,test,val
        return(im)


    def __getitem__(self, index):
        # set file paths
        file_name = self.files[index]
        file_path_hr = os.path.join(self.hr_path, file_name)
        file_path_lr = os.path.join(self.lr_path, file_name)
        
        # PIL
        hr = Image.open(file_path_hr)
        lr = Image.open(file_path_lr)
        
        hr = np.array(hr)
        lr = np.array(lr)
        
        
        # change to no of bands wanted
        hr = hr[:,:,:self.bands]
        lr = lr[:,:,:self.bands]
        
        # Convert image to tensor
        hr = torch.Tensor(hr) / 255.0
        lr = torch.Tensor(lr) / 255.0

        # apply blur
        lr = self.apply_augmentations(lr)

        # apply convention norm
        hr = self.linear_transform(hr,stage="norm")
        lr = self.linear_transform(lr,stage="norm")

        images = {"LR_image":lr,"image":hr}
        
        return images


"""
ds = s2_dataset(directory="/data2/simon/S2_images/",phase="train",bands=4)
dl = DataLoader(ds,batch_size=1,shuffle=True,num_workers=16,pin_memory=True)

from ldm.helper_functions.linear_transforms import linear_transform
from ldm.helper_functions.stretching import convention_stretch_sen2
import matplotlib.pyplot as plt
batch = next(iter(dl))
batch["LR_image"] = linear_transform(batch["LR_image"],stage="denorm")
batch["image"] = linear_transform(batch["image"],stage="denorm")
batch["LR_image"] = convention_stretch_sen2(batch["LR_image"])
batch["image"] = convention_stretch_sen2(batch["image"])
# plot HR and LR image from batch dictionary
plt.figure(figsize=(20,10))
plt.subplot(1,2,1)
plt.imshow(batch["image"][0][:,:,:3].cpu().numpy())
plt.title("HR image")
plt.subplot(1,2,2)
plt.imshow(batch["LR_image"][0][:,:,:3].cpu().numpy())
plt.title("LR image")
plt.savefig("S2_ds.png")
plt.close()
"""


import torch
import os
import pandas as pd
import torch
import random
from einops import rearrange
from torchvision import transforms
import numpy as np
from tqdm import tqdm
from torch.utils.data import Dataset, DataLoader, WeightedRandomSampler


# Step 2: Custom Dataset Class
class S2NAIP_v4(Dataset):
    def __init__(self, csv_path, phase="train",apply_norm=False):
        
        # set properties
        self.phase = phase
        assert phase in ["train","val","test"], "Phase must be either 'train', 'val' or 'test'"
        self.apply_norm=apply_norm

        # set paths
        self.input_dir = "/data3/final_s2naip_simon/"
        self.hr_input_path = os.path.join(self.input_dir,self.phase,"HR")
        self.lr_input_path = os.path.join(self.input_dir,self.phase,"LR")
        self.degradations = ["none","gaussian","bell","sigmoid"]

        # read and clean DF
        if self.phase=="train":
            self.dataframe = pd.read_csv(csv_path)
            self.dataframe = self.dataframe.fillna(0)
            self.dataframe = self.dataframe[self.dataframe["SuperClass"]!=0]
        if self.phase=="val" or self.phase=="test":
            dfs = [[],[],[],[],[]]
            # read all files of folder into pandas
            for filename in os.listdir(self.hr_input_path):
                file_path = os.path.join( self.hr_input_path, filename)
                # Check if it is a file
                if os.path.isfile(file_path):
                    # Assuming the files are CSVs. For Excel, use read_excel.
                    dfs[0].append(file_path)
                    dfs[1].append(file_path)
                    dfs[2].append(file_path)
                    dfs[3].append(file_path)
                    dfs[4].append(file_path)
            df = pd.DataFrame(dfs)
            df = df.T
            df.columns = ["HR","none","gaussian","bell","sigmoid"]
            self.dataframe = df

        # initialize sampler
        if self.phase=="train":
            self.define_sampler()
        # initialize norm
        from opensr_dataloaders.linear_transforms import linear_transform
        self.linear_transform = linear_transform
        # init albumentations
        from opensr_dataloaders.image_augmentations import init_color_transform
        self.color_transform = init_color_transform(brightness=0.5,contrast=0.5,saturation=0.5,hue=0.5)
        from opensr_dataloaders.image_augmentations import augment_bands
        self.augment_bands = augment_bands
        from opensr_dataloaders.image_augmentations import add_black_spots
        self.add_black_spots = add_black_spots
        from opensr_dataloaders.image_augmentations import other_augmentations
        self.other_augmentations = other_augmentations

    def __len__(self):
        return len(self.dataframe)
    
    def normalization(self, im, stage="norm"):
        im = im.unsqueeze(0)
        im = self.linear_transform(im,stage=stage)
        im = im.squeeze(0)
        return(im)
    
    def apply_augmentations(self,lr,hr):
        lr,hr = lr.float(),hr.float()

        # set probabilities
        smoothen_liklelihood   = 0.80 # 0.75
        jitter_liklelihood     = 0.40 # 0.75
        black_spots_likelihood = 0.00

        # get random numbers
        smoothen_rand = random.uniform(0, 1)
        jitter_rand = random.uniform(0, 1)
        black_spots_rand = random.uniform(0, 1)

        # perform color jitter
        if jitter_rand<jitter_liklelihood:
            lr,hr = self.augment_bands(lr,hr,self.color_transform)

        # perform other augmentations
        lr,hr = self.other_augmentations(lr,hr)

        # perform smoothen
        if smoothen_rand<smoothen_liklelihood:
            # Define Kernel
            sigma_rand = random.uniform(0.65, 1.2)
            gaussian_blur = transforms.GaussianBlur(kernel_size=5, sigma=sigma_rand)
            # Apply the blur to the image tensor
            band_ls = []
            for band in lr:
                band = torch.unsqueeze(band,0)
                band = gaussian_blur(band)
                band = torch.squeeze(band,0)
                band_ls.append(band)
            lr = torch.stack(band_ls)
        
        if black_spots_likelihood<black_spots_rand:
            lr = self.add_black_spots(lr)
            
                
        return(lr,hr)
    
    def define_sampler(self):
        # stratiefied Sampling with weights - ideally it sums up to 1
        class_weights = {
            'Rural': 0.30,  
            'Forest': 0.25, 
            'Water': 0.05,  
            'Developed': 0.4, }
        
        # 1. clean dataset
        data = self.dataframe
        # 2. define sampler
        class_counts = data['SuperClass'].value_counts()
        base_weights = [1.0 / class_counts[i] for i in data['SuperClass'].values]
        adjusted_weights = [base_weights[i] * class_weights[class_label] for i, class_label in enumerate(data['SuperClass'].values)] 
        # 3. set sampler
        self.sampler = WeightedRandomSampler(adjusted_weights, len(adjusted_weights),replacement=True)
    
    def __getitem__(self, idx):
        # get random degradation, get input paths
        random_degradation = np.random.choice(self.degradations)
        lr_input_path = os.path.join(self.input_dir,self.phase,"LR",random_degradation)

        if self.phase=="train":
            lr_input_path = os.path.join(lr_input_path,self.dataframe.iloc[idx]["name"] + ".pt")
            hr_input_path = os.path.join(self.hr_input_path,self.dataframe.iloc[idx]["name"] + ".pt")
        if self.phase=="val":
            hr_input_path = self.dataframe.iloc[[idx]]["HR"].item()
            lr_input_path = self.dataframe.iloc[[idx]][random_degradation].item()
        if self.phase=="test":
            hr_input_path = self.dataframe.iloc[[idx]]["HR"].item()
            lr_input_path = self.dataframe.iloc[[idx]][random_degradation].item()

        # check if files exist
        if not os.path.exists(lr_input_path) or not os.path.exists(hr_input_path):
            # if files arent yet on disk, load another random example
            #idx = np.random.randint(0,len(self))
            print("WARINING: file not available. Loading random...")
            #return self.__getitem__(idx) # recursive call
            pass

       # Load iamges from disk
        lr_image = torch.load(lr_input_path)
        hr_image = torch.load(hr_input_path)
        lr_image = lr_image.float()
        lr_image = lr_image.float()

        # bring to value range. Check since they are mixed while writing to disk is ongoing
        if hr_image.max()>10:
            hr_image = hr_image/10000
        if lr_image.max()>10:
            lr_image = lr_image/10000

        # perform augmentations
        if self.phase=="train":
            lr_image,hr_image = self.apply_augmentations(lr_image,hr_image)
            
        # apply norm
        apply_norm = True
        if self.apply_norm:
            lr_image,hr_image = lr_image.unsqueeze(0),hr_image.unsqueeze(0)
            lr_image = self.linear_transform(lr_image,stage="norm")
            hr_image = self.linear_transform(hr_image,stage="norm")
            lr_image,hr_image = lr_image.squeeze(0),hr_image.squeeze(0)

        hr_image = rearrange(hr_image,"c h w -> h w c")
        lr_image = rearrange(lr_image,"c h w -> h w c")

        # return images
        return {"image":hr_image,"LR_image":lr_image}

if False:
    ds = S2NAIP_v4(phase="val",csv_path="/data3/landcover_s2naip/csvs/train_metadata_landcover.csv")
    dl = DataLoader(ds,batch_size=1,shuffle=False,num_workers=0)
    lr_means = []
    hr_means = []
    for batch in tqdm(dl):
        lr_means.append(batch["LR_image"].max().item())
        hr_means.append(batch["image"].max().item())

        if batch["LR_image"].max().item()>10 or batch["image"].max().item()>10:
            print("Unusual: ",batch["LR_image"].max().item(),batch["image"].max().item())
    


if False:
    # create objects
    ds = S2NAIP_v4(csv_path="/data3/landcover_s2naip/csvs/train_metadata_landcover.csv")
    print("Lenght: ",len(ds))
    batch = ds.__getitem__(5679)
    hr,lr = batch["image"],batch["LR_image"]
    # show lr and hr images
    import matplotlib.pyplot as plt
    from ldm.helper_functions.stretching import convention_stretch_sen2
    fig,ax = plt.subplots(1,2,figsize=(20,10))
    lr_viz,hr_viz = lr,hr
    lr_viz = ds.linear_transform(lr_viz,stage="denorm")
    hr_viz = ds.linear_transform(hr_viz,stage="denorm")
    #lr_viz,hr_viz = (lr_viz+1)/2,(hr_viz+1)/2
    hr_viz = rearrange(hr_viz,"c h w -> h w c")
    lr_viz = rearrange(lr_viz,"c h w -> h w c")
    hr_viz = convention_stretch_sen2(hr_viz)
    lr_viz = convention_stretch_sen2(lr_viz)
    ax[0].imshow(hr_viz[:,:,:3])
    ax[1].imshow(lr_viz[:,:,:3])
    plt.savefig("test.png")







from setuptools import setup, find_packages

# read the contents of README file
with open('README.md', encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='opensr-dataloaders',
    version='0.0.11',
    author = "Simon Donike, Cesar Aybar, Luis Gomez Chova, Freddie Kalaitzis",
    author_email = "accounts@donike.net",
    description = "TEMPORARY: internal tool for dataset loading and sharing",
    url = "https://isp.uv.es/opensr/",
    project_urls={'Source Code': 'https://github.com/ESAopenSR/opensr-dataloaders'},
    license='MIT',
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
	'numpy==1.23.5',
	'einops==0.6.0',
	'rasterio==1.3.6',
	'tqdm==4.64.1',
	'torch==1.13.1',
	'scikit-image==0.19.3',
        'pytorch-lightning==1.9.0',
    	'torchvision==0.14.1',
    	'pandas==2.2.1'],
)



import torch
from einops import rearrange





def linear_transform(t_input,stage="norm"):
    assert stage in ["norm","denorm"]
    # get the shape of the tensor
    shape = t_input.shape

    # if 5 d tensor, norm/denorm individually
    if len(shape)==5:
        stack = []
        for batch in t_input:
            stack2 = []
            for i in range(0, t_input.size(1), 4):
                slice_tensor = batch[i:i+4, :, :, :]
                slice_denorm = linear_transform(slice_tensor,stage=stage)
                stack2.append(slice_denorm)
            stack2 = torch.stack(stack2)
            stack2 = stack2.reshape(shape[1], shape[2], shape[3], shape[4])
            stack.append(stack2)
        stack = torch.stack(stack)
        return(stack)

    # here only if len(shape) == 4
    squeeze_needed = False
    if len( shape ) == 3:
        squeeze_needed = True
        t_input = t_input.unsqueeze(0)
        shape = t_input.shape

    assert len(shape)==4 or len(shape)==5,"Input tensor must have 4 dimensions (B,C,H,W) - or 5D for MISR"
    transpose_needed = False
    if shape[-1]>shape[1]:
        transpose_needed = True
        t_input = rearrange(t_input,"b c h w -> b w h c")
    
    # define constants
    rgb_c = 3.
    nir_c = 5.

    # iterate over batches
    return_ls = []
    for t in t_input:
        if stage == "norm":
            # divide according to conventions
            t[:,:,0] = t[:,:,0] * (10.0 / rgb_c) # R
            t[:,:,1] = t[:,:,1] * (10.0 / rgb_c) # G
            t[:,:,2] = t[:,:,2] * (10.0 / rgb_c) # B
            t[:,:,3] = t[:,:,3] * (10.0 / nir_c) # NIR
            # clamp to get rif of outlier pixels
            t = t.clamp(0,1)
            # bring to -1..+1
            t = (t*2)-1
        if stage == "denorm":
            # bring to 0..1
            t = (t+1)/2
            # divide according to conventions
            t[:,:,0] = t[:,:,0] * (rgb_c / 10.0) # R
            t[:,:,1] = t[:,:,1] * (rgb_c / 10.0) # G
            t[:,:,2] = t[:,:,2] * (rgb_c / 10.0) # B
            t[:,:,3] = t[:,:,3] * (nir_c / 10.0) # NIR
            # clamp to get rif of outlier pixels
            t = t.clamp(0,1)
        
        # append result to list
        return_ls.append(t)

    # after loop, stack image
    t_output = torch.stack(return_ls)
    #print("stacked",t_output.shape)

    if transpose_needed==True:
        t_output = rearrange(t_output,"b w h c -> b c h w")
    if squeeze_needed:
        t_output = t_output.squeeze(0)

    return(t_output)




import torch
import numpy as np
import random
import albumentations as A
from einops import rearrange 


def init_color_transform(brightness=0.2,contrast=0.2,saturation=0.2,hue=0.2):
    transform = A.ReplayCompose([A.ColorJitter(brightness=brightness, contrast=contrast, saturation=saturation, hue=hue, always_apply=True)])
    return(transform)


def augment_bands(lr, hr, transform):
    """
    Applies augmentation to three randomly selected bands of 4-band (W, H, 4 or 6) tensors using Albumentations.

    Parameters:
    lr (torch.Tensor): Low-resolution input tensor.
    hr (torch.Tensor): High-resolution input tensor.
    transform (albumentations.core.composition.Compose): Albumentations transformation.

    Returns:
    Tuple[torch.Tensor, torch.Tensor]: Augmented low-resolution and high-resolution tensors.

    Note:
    - Assumes last dimension represents bands.
    - Converts tensors to/from NumPy arrays for augmentation.
    """
    
    # convert to Numpy
    lr = lr.numpy()
    hr = hr.numpy()

    rearrange_needed = False
    if lr.shape[-1]>lr.shape[0]:
        rearrange_needed = True
        lr = rearrange(lr,"c h w -> h w c")
        hr = rearrange(hr,"c h w -> h w c")

    assert lr.shape[-1]==4 or lr.shape[-1]==6,"Augmentation Shape invalid"
    assert hr.shape[-1]==4 or lr.shape[-1]==6,"Augmentation Shape invalid"
    
    # Select 3 random indices from the 4 bands
    bands = range(lr.shape[-1])
    selected_indices = random.sample(bands, 3)
    
    # Extract the selected bands
    lr_selected = lr[:, :, selected_indices]
    hr_selected = hr[:, :, selected_indices]
    
    # Apply augmentation
    lr_augmented = transform(image=lr_selected)
    hr_augmented = A.ReplayCompose.replay(lr_augmented['replay'], image=hr_selected)["image"]
    lr_augmented = lr_augmented["image"]

    # Reconstruct the original 4-band tensor
    lr[:, :, selected_indices] = lr_augmented
    hr[:, :, selected_indices] = hr_augmented

    if rearrange_needed==True:
        lr = rearrange(lr,"h w c -> c h w")
        hr = rearrange(hr,"h w c -> c h w")
    
    # convert back to Tensor
    lr = torch.Tensor(lr)
    hr = torch.Tensor(hr)

    return lr, hr




def add_black_spots(images,mean_amount=15,mean_size=8):
    """
    Add black spots of varying sizes and amounts to all bands of the input image.
    
    Parameters:
    - image: A PyTorch tensor of shape (batch,channels, height, width).
    - mean_amount: mean of notm. distribution for amount of black spots
    - mean_size: mean of norm. distribution of black spot's size
    
    Returns:
    - Modified image with black spots.
    """
    
    
    if len(images.shape)==3:
        images = images.unsqueeze(0)
        three_band = False
    else:
        three_band = True
    
    final_result = []
    for image in images:
        # get random number of spots form normal distribution
        mean = mean_amount
        std = 3  # Adjust this as needed
        num_spots_f = torch.normal(mean, std, size=(1,))
        num_spots = max(0, round(num_spots_f.item()))

        channels, height, width = image.shape
        for _ in range(num_spots):
            # Randomly determine the center of the spot
            center_x, center_y = torch.randint(0, width, (1,)), torch.randint(0, height, (1,))

            # Randomly determine the size of the spot
            std_size = 3
            rand_size = torch.normal(mean_size, std_size, size=(1,))
            rand_size = max(0, round(rand_size.item()))
            spot_size = rand_size

            # Calculate the top-left corner of the spot
            start_x = max(center_x - spot_size // 2, 0)
            start_y = max(center_y - spot_size // 2, 0)

            # Calculate the bottom-right corner of the spot
            end_x = min(start_x + spot_size, width)
            end_y = min(start_y + spot_size, height)

            # Set the pixels in this range to 0 in all bands
            image[:, start_y:end_y, start_x:end_x] = 0.0
        
        # append image with spots to list
        final_result.append(image)
    
    # stack result
    final_result = torch.stack(final_result)
    
    # back to 3D tensor if necessary
    if three_band == False:
        final_result = final_result.squeeze(0)
    return final_result



def other_augmentations(lr,hr):
    """
    Applies identical augmentations including HorizontalFlip, VerticalFlip, RandomRotate90, 
    and Transpose to all bands of both low-resolution and high-resolution images using Albumentations.
    This ensures the spatial correspondence between LR and HR images is maintained.

    Parameters:
    lr (torch.Tensor): Low-resolution input tensor.
    hr (torch.Tensor): High-resolution input tensor.

    Returns:
    Tuple[torch.Tensor, torch.Tensor]: Augmented low-resolution and high-resolution tensors.

    Note:
    - Assumes last dimension represents bands.
    - Converts tensors to/from NumPy arrays for augmentation.
    """
    
    # Define the transformation pipeline
    transform = A.ReplayCompose([
        A.HorizontalFlip(p=0.5),
        A.VerticalFlip(p=0.5),
        A.RandomRotate90(p=0.5),
        A.Transpose(p=0.5)
    ])
    
    # Convert tensors to numpy arrays
    lr_np = lr.numpy()
    hr_np = hr.numpy()

    # Check and rearrange if channels are first
    rearrange_needed = False
    if lr_np.shape[-1] > lr_np.shape[0]:
        rearrange_needed = True
        lr_np = rearrange(lr_np, "c h w -> h w c")
        hr_np = rearrange(hr_np, "c h w -> h w c")

    # Apply augmentation to the LR image and get the replay data
    augmented_lr = transform(image=lr_np)
    replay_data = augmented_lr['replay']

    # Apply the same augmentation to the HR image using the replay data
    augmented_hr = A.ReplayCompose.replay(replay_data, image=hr_np)

    # Extract augmented images
    lr_np = augmented_lr['image']
    hr_np = augmented_hr['image']

    # Rearrange back if needed
    if rearrange_needed:
        lr_np = rearrange(lr_np, "h w c -> c h w")
        hr_np = rearrange(hr_np, "h w c -> c h w")
    
    # Convert back to PyTorch tensors
    lr_tensor = torch.from_numpy(lr_np).float()
    hr_tensor = torch.from_numpy(hr_np).float()

    return lr_tensor, hr_tensor